# Generated by Django 3.1 on 2021-11-29 10:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security_backends_sql', '0002_auto_20211119_1940'),
    ]

    operations = [
        migrations.AddField(
            model_name='celerytaskinvocationlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='celerytaskrunlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='commandlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='inputrequestlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='outputrequestlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
    ]
